@@ -55,8 +55,7 @@ module Database.PostgreSQL.Simple
     -- $types
 
     -- * Types
-      Base.ConnectInfo(..)
-    , Connection
+      Connection
     , Query
     , ToRow
     , FromRow
@@ -67,15 +66,16 @@ module Database.PostgreSQL.Simple
     -- ** Exceptions
     , SqlError(..)
     , PQ.ExecStatus(..)
-    , FormatError(fmtMessage, fmtQuery, fmtParams)
-    , QueryError(qeMessage, qeQuery)
-    , ResultError(errSQLType, errHaskellType, errMessage)
+    , FormatError(..)
+    , QueryError(..)
+    , ResultError(..)
     -- * Connection management
-    , Base.connect
     , Base.connectPostgreSQL
-    , Base.postgreSQLConnectionString
-    , Base.defaultConnectInfo
     , Base.close
+    , Base.connect
+    , Base.ConnectInfo(..)
+    , Base.defaultConnectInfo
+    , Base.postgreSQLConnectionString
     -- * Queries that return results
     , query
     , query_
@@ -114,19 +114,18 @@ import           Blaze.ByteString.Builder
                    ( Builder, fromByteString, toByteString )
 import           Blaze.ByteString.Builder.Char8 (fromChar)
 import           Blaze.Text ( integral )
-import           Control.Applicative ((<$>), pure)
+import           Control.Applicative ((<$>))
 import           Control.Exception as E
 import           Control.Monad (foldM)
 import           Data.ByteString (ByteString)
 import           Data.Int (Int64)
 import           Data.List (intersperse)
 import           Data.Monoid (mconcat)
-import           Data.Typeable (Typeable)
 import           Database.PostgreSQL.Simple.Compat ( (<>) )
 import           Database.PostgreSQL.Simple.FromField (ResultError(..))
 import           Database.PostgreSQL.Simple.FromRow (FromRow(..))
 import           Database.PostgreSQL.Simple.Ok
-import           Database.PostgreSQL.Simple.ToField (Action(..), inQuotes)
+import           Database.PostgreSQL.Simple.ToField (Action(..))
 import           Database.PostgreSQL.Simple.ToRow (ToRow(..))
 import           Database.PostgreSQL.Simple.Types
                    ( Binary(..), In(..), Only(..), Query(..), (:.)(..) )
@@ -135,21 +134,9 @@ import           Database.PostgreSQL.Simple.Transaction
 import           Database.PostgreSQL.Simple.TypeInfo
 import qualified Database.PostgreSQL.LibPQ as PQ
 import qualified Data.ByteString.Char8 as B
-import qualified Data.Text          as T
-import qualified Data.Text.Encoding as TE
 import           Control.Monad.Trans.Reader
 import           Control.Monad.Trans.State.Strict
 
--- | Exception thrown if a 'Query' could not be formatted correctly.
--- This may occur if the number of \'@?@\' characters in the query
--- string does not match the number of parameters provided.
-data FormatError = FormatError {
-      fmtMessage :: String
-    , fmtQuery :: Query
-    , fmtParams :: [ByteString]
-    } deriving (Eq, Show, Typeable)
-
-instance Exception FormatError
 
 -- | Format a query string.
 --
@@ -286,31 +273,17 @@ parseTemplate template =
 
     skipSpace = B.dropWhile isSpace_ascii
 
-escapeStringConn :: Connection -> ByteString -> IO (Either ByteString ByteString)
-escapeStringConn conn s =
-    withConnection conn $ \c ->
-    PQ.escapeStringConn c s >>= checkError c
 
-escapeByteaConn :: Connection -> ByteString -> IO (Either ByteString ByteString)
-escapeByteaConn conn s =
-    withConnection conn $ \c ->
-    PQ.escapeByteaConn c s >>= checkError c
-
-checkError :: PQ.Connection -> Maybe a -> IO (Either ByteString a)
-checkError _ (Just x) = return $ Right x
-checkError c Nothing  = Left . maybe "" id <$> PQ.errorMessage c
 
 buildQuery :: Connection -> Query -> ByteString -> [Action] -> IO Builder
-buildQuery conn q template xs = zipParams (split template) <$> mapM sub xs
-  where quote = either (\msg -> fmtError (utf8ToString msg) q xs)
-                       (inQuotes . fromByteString)
-        utf8ToString = T.unpack . TE.decodeUtf8
-        sub (Plain  b)      = pure b
-        sub (Escape s)      = quote <$> escapeStringConn conn s
-        sub (EscapeByteA s) = quote <$> escapeByteaConn conn s
-        sub (Many  ys)      = mconcat <$> mapM sub ys
-        split s = fromByteString h : if B.null t then [] else split (B.tail t)
-            where (h,t) = B.break (=='?') s
+buildQuery conn q template xs =
+    zipParams (split template) <$> mapM (buildAction conn q xs) xs
+  where split s =
+            let (h,t) = B.break (=='?') s
+            in fromByteString h
+               : if B.null t
+                 then []
+                 else split (B.tail t)
         zipParams (t:ts) (p:ps) = t <> p <> zipParams ts ps
         zipParams [t] []        = t
         zipParams _ _ = fmtError (show (B.count '?' template) ++
@@ -322,7 +295,8 @@ buildQuery conn q template xs = zipParams (split template) <$> mapM sub xs
 --
 -- Returns the number of rows affected.
 --
--- Throws 'FormatError' if the query could not be formatted correctly.
+-- Throws 'FormatError' if the query could not be formatted correctly, or
+-- a 'SqlError' exception if the backend returns an error.
 execute :: (ToRow q) => Connection -> Query -> q -> IO Int64
 execute conn template qs = do
   result <- exec conn =<< formatQuery conn template qs
@@ -331,9 +305,33 @@ execute conn template qs = do
 -- | Execute a multi-row @INSERT@, @UPDATE@, or other SQL query that is not
 -- expected to return results.
 --
--- Returns the number of rows affected.
+-- Returns the number of rows affected.   If the list of parameters is empty,
+-- this function will simply return 0 without issuing the query to the backend.
+-- If this is not desired, consider using the 'Values' constructor instead.
 --
--- Throws 'FormatError' if the query could not be formatted correctly.
+-- Throws 'FormatError' if the query could not be formatted correctly, or
+-- a 'SqlError' exception if the backend returns an error.
+--
+-- For example,  here's a command that inserts two rows into a table
+-- with two columns:
+--
+-- @
+-- executeMany c [sql|
+--     INSERT INTO sometable VALUES (?,?)
+--  |] [(1, \"hello\"),(2, \"world\")]
+-- @
+--
+-- Here's an canonical example of a multi-row update command:
+--
+-- @
+-- executeMany c [sql|
+--     UPDATE sometable
+--        SET sometable.y = upd.y
+--       FROM (VALUES (?,?)) as upd(x,y)
+--      WHERE sometable.x = upd.x
+--  |] [(1, \"hello\"),(2, \"world\")
+-- @
+
 executeMany :: (ToRow q) => Connection -> Query -> [q] -> IO Int64
 executeMany _ _ [] = return 0
 executeMany conn q qs = do
@@ -347,6 +345,10 @@ executeMany conn q qs = do
 -- in cases where you are only inserting a single row,  and do not need
 -- functionality analogous to 'executeMany'.
 --
+-- If the list of parameters is empty,  this function will simply return @[]@
+-- without issuing the query to the backend.   If this is not desired,
+-- consider using the 'Values' constructor instead.
+--
 -- Throws 'FormatError' if the query could not be formatted correctly.
 returning :: (ToRow q, FromRow r) => Connection -> Query -> [q] -> IO [r]
 returning _ _ [] = return []
@@ -369,6 +371,9 @@ returning conn q qs = do
 --   using 'execute' instead of 'query').
 --
 -- * 'ResultError': result conversion failed.
+--
+-- * 'SqlError':  the postgresql backend returned an error,  e.g.
+--   a syntax or type error,  or an incorrect table or column name.
 query :: (ToRow q, FromRow r) => Connection -> Query -> q -> IO [r]
 query = queryWith fromRow
 
@@ -398,6 +403,12 @@ queryWith_ parser conn q@(Query que) = do
 -- This fold is /not/ strict. The stream consumer is responsible for
 -- forcing the evaluation of its result to avoid space leaks.
 --
+-- This is implemented using a database cursor.    As such,  this requires
+-- a transaction.   This function will detect whether or not there is a
+-- transaction in progress,  and will create a 'ReadCommitted' 'ReadOnly'
+-- transaction if needed.   The cursor is given a unique temporary name,
+-- so the consumer may itself call fold.
+--
 -- Exceptions that may be thrown:
 --
 -- * 'FormatError': the query string could not be formatted correctly.
@@ -406,7 +417,9 @@ queryWith_ parser conn q@(Query que) = do
 --   using 'execute' instead of 'query').
 --
 -- * 'ResultError': result conversion failed.
-
+--
+-- * 'SqlError':  the postgresql backend returned an error,  e.g.
+--   a syntax or type error,  or an incorrect table or column name.
 fold            :: ( FromRow row, ToRow params )
                 => Connection
                 -> Query
@@ -416,6 +429,9 @@ fold            :: ( FromRow row, ToRow params )
                 -> IO a
 fold = foldWithOptions defaultFoldOptions
 
+-- | Number of rows to fetch at a time.   'Automatic' currently defaults
+--   to 256 rows,  although it might be nice to make this more intelligent
+--   based on e.g. the average size of the rows.
 data FetchQuantity
    = Automatic
    | Fixed !Int
@@ -426,12 +442,19 @@ data FoldOptions
        transactionMode :: !TransactionMode
      }
 
+-- | defaults to 'Automatic',  and 'TransactionMode' 'ReadCommitted' 'ReadOnly'
 defaultFoldOptions :: FoldOptions
 defaultFoldOptions = FoldOptions {
       fetchQuantity   = Automatic,
       transactionMode = TransactionMode ReadCommitted ReadOnly
     }
 
+-- | The same as 'fold',  but this provides a bit more control over
+--   lower-level details.  Currently,  the number of rows fetched per
+--   round-trip to the server and the transaction mode may be adjusted
+--   accordingly.    If the connection is already in a transaction,
+--   then the existing transaction is used and thus the 'transactionMode'
+--   option is ignored.
 foldWithOptions :: ( FromRow row, ToRow params )
                 => FoldOptions
                 -> Connection
@@ -600,16 +623,6 @@ ellipsis bs
     | B.length bs > 15 = B.take 10 bs `B.append` "[...]"
     | otherwise        = bs
 
-fmtError :: String -> Query -> [Action] -> a
-fmtError msg q xs = throw FormatError {
-                      fmtMessage = msg
-                    , fmtQuery = q
-                    , fmtParams = map twiddle xs
-                    }
-  where twiddle (Plain b)       = toByteString b
-        twiddle (Escape s)      = s
-        twiddle (EscapeByteA s) = s
-        twiddle (Many ys)       = B.concat (map twiddle ys)
 
 -- $use
 --
@@ -637,9 +650,11 @@ fmtError msg q xs = throw FormatError {
 -- >
 -- > import Database.PostgreSQL.Simple
 -- >
+-- > hello :: IO Int
 -- > hello = do
--- >   conn <- connect defaultConnectInfo
--- >   query_ conn "select 2 + 2"
+-- >   conn <- connectPostgreSQL ""
+-- >   [Only i] <- query_ conn "select 2 + 2"
+-- >   return i
 --
 -- A 'Query' value does not represent the actual query that will be
 -- executed, but is a template for constructing the final query.
@@ -700,11 +715,26 @@ fmtError msg q xs = throw FormatError {
 -- The same kind of problem can arise with string literals if you have
 -- the @OverloadedStrings@ language extension enabled.  Again, just
 -- use an explicit type signature if this happens.
+--
+-- Finally, remember that the compiler must be able to infer the type
+-- of a query's /results/ as well as its parameters.  We might like
+-- the following example to work:
+--
+-- > print =<< query_ conn "select 2 + 2"
+--
+-- Unfortunately, while a quick glance tells us that the result type
+-- should be a single row containing a single numeric column, the
+-- compiler has no way to infer what the types are.  We can easily fix
+-- this by providing an explicit type annotation:
+--
+-- > xs <- query_ conn "select 2 + 2"
+-- > print (xs :: [Only Int])
 
 -- $only_param
 --
 -- Haskell lacks a single-element tuple type, so if you have just one
--- value you want substituted into a query, what should you do?
+-- value you want substituted into a query or a single-column result,
+-- what should you do?
 --
 -- The obvious approach would appear to be something like this:
 --
@@ -719,6 +749,9 @@ fmtError msg q xs = throw FormatError {
 --
 -- > execute conn "insert into users (first_name) values (?)"
 -- >              ["Nuala"]
+--
+-- A row of /n/ query results is represented using an /n/-tuple, so
+-- you should use 'Only' to represent a single-column result.
 
 -- $in
 --
@@ -736,7 +769,7 @@ fmtError msg q xs = throw FormatError {
 -- Here's an example:
 --
 -- > query conn "select * from users where first_name in ?" $
--- >       In ["Anna", "Boris", "Carla"]
+-- >       Only $ In ["Anna", "Boris", "Carla"]
 --
 -- If your 'In'-wrapped list is empty, the string @\"(null)\"@ will be
 -- substituted instead, to ensure that your clause remains
