@@ -23,6 +23,24 @@ the Haskell 'Double' type is compatible with the PostgreSQL's 32-bit
 since a 'Double' might lose precision if representing PostgreSQL's 64-bit
 @bigint@, the two are /not/ considered compatible.
 
+Note that the 'Float' and 'Double' instances use attoparsec's 'double'
+conversion routine,  which sacrifices some accuracy for speed.   If you
+need accuracy,  consider first converting data to a 'Scientific' or 'Rational'
+type,  and then converting to a floating-point type.   If you are defining
+your own 'Database.PostgreSQL.Simple.FromRow.FromRow' instances,  this can be
+acheived simply by
+@'fromRational' '<$>' 'Database.PostgreSQL.Simple.FromRow.field'@,  although
+this idiom is additionally compatible with PostgreSQL's @numeric@ type.
+If this is unacceptable,  you may find
+'Database.PostgreSQL.Simple.FromRow.fieldWith' useful.
+
+Also note that while converting to a 'Double' through the 'Scientific' type
+is likely somewhat faster than converting through the 'Rational' type,
+the 'Scientific' type has no way to represent @NaN@ and @±Infinity@ values.
+Thus,  if you need precision conversion of regular floating point values
+and the possibility of receiving these special values from the backend,
+stick with 'Rational'.
+
 Because 'FromField' is a typeclass,  one may provide conversions to
 additional Haskell types without modifying postgresql-simple.  This is
 particularly useful for supporting PostgreSQL types that postgresql-simple
@@ -87,18 +105,18 @@ module Database.PostgreSQL.Simple.FromField
     , typeOid
     , PQ.Oid(..)
     , PQ.Format(..)
+    , pgArrayFieldParser
 
     , fromJSONField
     ) where
 
 #include "MachDeps.h"
 
-import           Control.Applicative
-                   ( Applicative, (<|>), (<$>), pure )
+import           Control.Applicative ( (<|>), (<$>), pure, (*>) )
 import           Control.Concurrent.MVar (MVar, newMVar)
 import           Control.Exception (Exception)
 import qualified Data.Aeson as JSON
-import           Data.Attoparsec.Char8 hiding (Result)
+import           Data.Attoparsec.ByteString.Char8 hiding (Result)
 import           Data.ByteString (ByteString)
 import qualified Data.ByteString.Char8 as B
 import           Data.Int (Int16, Int32, Int64)
@@ -125,8 +143,12 @@ import qualified Data.ByteString.Lazy as LB
 import qualified Data.Text as ST
 import qualified Data.Text.Encoding as ST
 import qualified Data.Text.Lazy as LT
+import           Data.CaseInsensitive (CI)
+import qualified Data.CaseInsensitive as CI
 import           Data.UUID   (UUID)
 import qualified Data.UUID as UUID
+import           Data.Scientific (Scientific)
+import           GHC.Real (infinity, notANumber)
 
 -- | Exception thrown if conversion from a SQL value to a Haskell
 -- value fails.
@@ -300,24 +322,45 @@ instance FromField Int64 where
 instance FromField Integer where
     fromField = atto ok64 $ signed decimal
 
--- | int2, float4
+-- | int2, float4    (Uses attoparsec's 'double' routine,  for
+--   better accuracy convert to 'Scientific' or 'Rational' first)
 instance FromField Float where
-    fromField = atto ok (realToFrac <$> double)
+    fromField = atto ok (realToFrac <$> pg_double)
       where ok = $(mkCompats [TI.float4,TI.int2])
 
--- | int2, int4, float4, float8
+-- | int2, int4, float4, float8  (Uses attoparsec's 'double' routine,  for
+--   better accuracy convert to 'Scientific' or 'Rational' first)
 instance FromField Double where
-    fromField = atto ok double
+    fromField = atto ok pg_double
       where ok = $(mkCompats [TI.float4,TI.float8,TI.int2,TI.int4])
 
 -- | int2, int4, float4, float8, numeric
 instance FromField (Ratio Integer) where
-    fromField = atto ok rational
+    fromField = atto ok pg_rational
+      where ok = $(mkCompats [TI.float4,TI.float8,TI.int2,TI.int4,TI.numeric])
+
+-- | int2, int4, float4, float8, numeric
+instance FromField Scientific where
+     fromField = atto ok rational
       where ok = $(mkCompats [TI.float4,TI.float8,TI.int2,TI.int4,TI.numeric])
 
 unBinary :: Binary t -> t
 unBinary (Binary x) = x
 
+pg_double :: Parser Double
+pg_double
+    =   (string "NaN"       *> pure ( 0 / 0))
+    <|> (string "Infinity"  *> pure ( 1 / 0))
+    <|> (string "-Infinity" *> pure (-1 / 0))
+    <|> double
+
+pg_rational :: Parser Rational
+pg_rational
+    =   (string "NaN"       *> pure notANumber )
+    <|> (string "Infinity"  *> pure infinity   )
+    <|> (string "-Infinity" *> pure (-infinity))
+    <|> rational
+
 -- | bytea, name, text, \"char\", bpchar, varchar, unknown
 instance FromField SB.ByteString where
     fromField f dat = if typeOid f == $(inlineTypoid TI.bytea)
@@ -357,6 +400,28 @@ instance FromField ST.Text where
 instance FromField LT.Text where
     fromField f dat = LT.fromStrict <$> fromField f dat
 
+-- | citext
+instance FromField (CI ST.Text) where
+    fromField f mdat = do
+       typ <- typename f
+       if typ /= "citext"
+         then returnError Incompatible f ""
+         else case mdat of
+                Nothing  -> returnError UnexpectedNull f ""
+                Just dat -> either left (pure . CI.mk)
+                                        (ST.decodeUtf8' dat)
+
+-- | citext
+instance FromField (CI LT.Text) where
+    fromField f mdat = do
+       typ <- typename f
+       if typ /= "citext"
+         then returnError Incompatible f ""
+         else case mdat of
+                Nothing  -> returnError UnexpectedNull f ""
+                Just dat -> either left (pure . CI.mk . LT.fromStrict)
+                                        (ST.decodeUtf8' dat)
+
 -- | name, text, \"char\", bpchar, varchar
 instance FromField [Char] where
     fromField f dat = ST.unpack <$> fromField f dat
@@ -426,27 +491,33 @@ instance (FromField a, FromField b) => FromField (Either a b) where
 
 -- | any postgresql array whose elements are compatible with type @a@
 instance (FromField a, Typeable a) => FromField (PGArray a) where
-    fromField f mdat = do
+  fromField = pgArrayFieldParser fromField
+
+pgArrayFieldParser :: Typeable a => FieldParser a -> FieldParser (PGArray a)
+pgArrayFieldParser fieldParser f mdat = do
         info <- typeInfo f
         case info of
           TI.Array{} ->
               case mdat of
                 Nothing  -> returnError UnexpectedNull f ""
                 Just dat -> do
-                   case parseOnly (fromArray info f) dat of
+                   case parseOnly (fromArray fieldParser info f) dat of
                      Left  err  -> returnError ConversionFailed f err
                      Right conv -> PGArray <$> conv
           _ -> returnError Incompatible f ""
 
-fromArray :: (FromField a)
-          => TypeInfo -> Field -> Parser (Conversion [a])
-fromArray typeInfo f = sequence . (parseIt <$>) <$> array delim
+fromArray :: FieldParser a -> TypeInfo -> Field -> Parser (Conversion [a])
+fromArray fieldParser typeInfo f = sequence . (parseIt <$>) <$> array delim
   where
     delim = typdelim (typelem typeInfo)
     fElem = f{ typeOid = typoid (typelem typeInfo) }
-    parseIt item = (fromField f' . Just . fmt delim) item
-      where f' | Arrays.Array _ <- item = f
-               | otherwise              = fElem
+
+    parseIt item =
+        fieldParser f' $ if item' == "NULL" then Nothing else Just item'
+      where
+        item' = fmt delim item
+        f' | Arrays.Array _ <- item = f
+           | otherwise              = fElem
 
 instance (FromField a, Typeable a) => FromField (Vector a) where
     fromField f v = V.fromList . fromPGArray <$> fromField f v
@@ -469,7 +540,7 @@ instance FromField UUID where
 -- | json
 instance FromField JSON.Value where
     fromField f mbs =
-      if typeOid f /= $(inlineTypoid TI.json)
+      if typeOid f /= $(inlineTypoid TI.json) && typeOid f /= $(inlineTypoid TI.jsonb)
       then returnError Incompatible f ""
       else case mbs of
              Nothing -> returnError UnexpectedNull f ""
