@@ -25,7 +25,7 @@ import Data.ByteString.Internal (c2w, w2c)
 import Data.Time hiding (getTimeZone, getZonedTime)
 import Data.Typeable
 import Data.Word(Word8)
-import qualified Data.Attoparsec.Char8 as A
+import qualified Data.Attoparsec.ByteString.Char8 as A
 import Data.Monoid(Monoid(..))
 import Data.Fixed (Pico)
 import Unsafe.Coerce
@@ -98,7 +98,7 @@ getDay = do
 
     let !year = toNum yearStr
     _       <- A.char '-'
-    month   <- digits "month"
+    !month  <- digits "month"
     _       <- A.char '-'
     day     <- digits "day"
 
@@ -115,7 +115,7 @@ decimal str = toNum str / 10^(B.length str)
 
 getTimeOfDay :: A.Parser TimeOfDay
 getTimeOfDay = do
-    hour   <- digits "hours"
+    !hour  <- digits "hours"
     _      <- A.char ':'
     minute <- digits "minutes"
     _      <- A.char ':'
@@ -136,13 +136,43 @@ getLocalTimestamp = getUnbounded getLocalTime
 
 getTimeZone :: A.Parser TimeZone
 getTimeZone = do
-    sign  <- A.satisfy (\c -> c == '+' || c == '-')
-    hours <- digits "timezone"
-    mins  <- (A.char ':' *> digits "timezone minutes") <|> pure 0
+    !sign  <- A.satisfy (\c -> c == '+' || c == '-')
+    !hours <- digits "timezone"
+    !mins  <- (A.char ':' *> digits "timezone minutes") <|> pure 0
     let !absset = 60 * hours + mins
         !offset = if sign == '+' then absset else -absset
     return $! minutesToTimeZone offset
 
+type TimeZoneHMS = (Int,Int,Int)
+
+getTimeZoneHMS :: A.Parser TimeZoneHMS
+getTimeZoneHMS = do
+    !sign  <- A.satisfy (\c -> c == '+' || c == '-')
+    !hours <- digits "timezone"
+    !mins  <- (A.char ':' *> digits "timezone minutes") <|> pure 0
+    !secs  <- (A.char ':' *> digits "timezone seconds") <|> pure 0
+    if sign == '+'
+    then return $! (hours, mins, secs)
+    else return $! (\ !h !m !s -> (h,m,s)) (-hours) (-mins) (-secs)
+
+localToUTCTimeOfDayHMS :: TimeZoneHMS -> TimeOfDay -> (Integer, TimeOfDay)
+localToUTCTimeOfDayHMS (dh, dm, ds) (TimeOfDay h m s) =
+    (\ !a !b -> (a,b)) dday (TimeOfDay h'' m'' s'')
+  where
+    s' = s - fromIntegral ds
+    (!s'', m')
+        | s' < 0    = (s' + 60, m - dm - 1)
+        | s' >= 60  = (s' - 60, m - dm + 1)
+        | otherwise = (s'     , m - dm    )
+    (!m'', h')
+        | m' < 0    = (m' + 60, h - dh - 1)
+        | m' >= 60  = (m' - 60, h - dh + 1)
+        | otherwise = (m'     , h - dh    )
+    (!h'', dday)
+        | h' < 0    = (h' + 24, -1)
+        | h' >= 24  = (h' - 24,  1)
+        | otherwise = (h'     ,  0)
+
 getZonedTime :: A.Parser ZonedTime
 getZonedTime = ZonedTime <$> getLocalTime <*> getTimeZone
 
@@ -154,11 +184,11 @@ getUTCTime = do
     day  <- getDay
     _    <- A.char ' '
     time <- getTimeOfDay
-    zone <- getTimeZone
-    let (!dayDelta,!time') = localToUTCTimeOfDay zone time
+    zone <- getTimeZoneHMS
+    let !(dayDelta,time') = localToUTCTimeOfDayHMS zone time
     let !day' = addDays dayDelta day
     let !time'' = timeOfDayToTime time'
-    return (UTCTime day' time'')
+    return $! UTCTime day' time''
 
 getUTCTimestamp :: A.Parser UTCTimestamp
 getUTCTimestamp = getUnbounded getUTCTime
@@ -229,6 +259,17 @@ localTimestampToBuilder = unboundedToBuilder localTimeToBuilder
 dateToBuilder  :: Date -> Builder
 dateToBuilder  = unboundedToBuilder dayToBuilder
 
+nominalDiffTimeToBuilder :: NominalDiffTime -> Builder
+nominalDiffTimeToBuilder xyz
+    | yz < 500000 = sign ++ integral x
+    | otherwise   = sign ++ integral x ++ fromChar '.' ++  showD6 y
+  where
+    -- A kludge to work around the fact that Data.Fixed isn't very fast and
+    -- doesn't give me access to the MkFixed constructor.
+    sign = if xyz >= 0 then mempty else fromChar '-'
+    (x,yz) = ((unsafeCoerce (abs xyz) :: Integer) + 500000)  `quotRem` 1000000000000
+    (fromIntegral -> y, _z) = yz `quotRem` 1000000
+
 showSeconds :: Pico -> Builder
 showSeconds xyz
     | yz == 0   = pad2 x
