@@ -1,4 +1,4 @@
-{-# LANGUAGE DeriveDataTypeable, DeriveFunctor       #-}
+{-# LANGUAGE CPP, DeriveDataTypeable, DeriveFunctor  #-}
 {-# LANGUAGE FlexibleInstances, TypeSynonymInstances #-}
 
 ------------------------------------------------------------------------------
@@ -30,9 +30,10 @@ import Data.ByteString (ByteString)
 import Data.Int (Int8, Int16, Int32, Int64)
 import Data.List (intersperse)
 import Data.Monoid (mappend)
-import Data.Time (Day, TimeOfDay, LocalTime, UTCTime, ZonedTime)
+import Data.Time (Day, TimeOfDay, LocalTime, UTCTime, ZonedTime, NominalDiffTime)
 import Data.Typeable (Typeable)
 import Data.Word (Word, Word8, Word16, Word32, Word64)
+import {-# SOURCE #-} Database.PostgreSQL.Simple.ToRow
 import Database.PostgreSQL.Simple.Types
 import qualified Blaze.ByteString.Builder.Char.Utf8 as Utf8
 import qualified Data.ByteString as SB
@@ -40,12 +41,19 @@ import qualified Data.ByteString.Lazy as LB
 import qualified Data.Text as ST
 import qualified Data.Text.Encoding as ST
 import qualified Data.Text.Lazy as LT
+import qualified Data.Text.Lazy.Builder as LT
 import           Data.UUID   (UUID)
 import qualified Data.UUID as UUID
 import           Data.Vector (Vector)
 import qualified Data.Vector as V
 import qualified Database.PostgreSQL.LibPQ as PQ
 import           Database.PostgreSQL.Simple.Time
+import           Data.Scientific (Scientific)
+#if MIN_VERSION_scientific(0,3,0)
+import           Data.Text.Lazy.Builder.Scientific (scientificBuilder)
+#else
+import           Data.Scientific (scientificBuilder)
+#endif
 
 -- | How to render an element when substituting it into a query.
 data Action =
@@ -61,15 +69,20 @@ data Action =
   | EscapeByteA ByteString
     -- ^ Escape binary data for use as a @bytea@ literal.  Include surrounding
     -- quotes.  This is used by the 'Binary' newtype wrapper.
+  | EscapeIdentifier ByteString
+    -- ^ Escape before substituting. Use for all sql identifiers like
+    -- table, column names, etc. This is used by the 'Identifier' newtype
+    -- wrapper.
   | Many [Action]
     -- ^ Concatenate a series of rendering actions.
     deriving (Typeable)
 
 instance Show Action where
-    show (Plain b)       = "Plain " ++ show (toByteString b)
-    show (Escape b)      = "Escape " ++ show b
-    show (EscapeByteA b) = "EscapeByteA " ++ show b
-    show (Many b)        = "Many " ++ show b
+    show (Plain b)            = "Plain " ++ show (toByteString b)
+    show (Escape b)           = "Escape " ++ show b
+    show (EscapeByteA b)      = "EscapeByteA " ++ show b
+    show (EscapeIdentifier b) = "EscapeIdentifier " ++ show b
+    show (Many b)             = "Many " ++ show b
 
 -- | A type that may be used as a single parameter to a SQL query.
 class ToField a where
@@ -166,6 +179,10 @@ instance ToField Double where
               | otherwise               = Plain (double v)
     {-# INLINE toField #-}
 
+instance ToField Scientific where
+    toField x = toField (LT.toLazyText (scientificBuilder x))
+    {-# INLINE toField #-}
+
 instance ToField (Binary SB.ByteString) where
     toField (Binary bs) = EscapeByteA bs
     {-# INLINE toField #-}
@@ -174,6 +191,20 @@ instance ToField (Binary LB.ByteString) where
     toField (Binary bs) = (EscapeByteA . SB.concat . LB.toChunks) bs
     {-# INLINE toField #-}
 
+instance ToField Identifier where
+    toField (Identifier bs) = EscapeIdentifier (ST.encodeUtf8 bs)
+    {-# INLINE toField #-}
+
+instance ToField QualifiedIdentifier where
+    toField (QualifiedIdentifier (Just s) t) =
+        Many [ EscapeIdentifier (ST.encodeUtf8 s)
+             , Plain (fromChar '.')
+             , EscapeIdentifier (ST.encodeUtf8 t)
+             ]
+    toField (QualifiedIdentifier Nothing  t) =
+               EscapeIdentifier (ST.encodeUtf8 t)
+    {-# INLINE toField #-}
+
 instance ToField SB.ByteString where
     toField = Escape
     {-# INLINE toField #-}
@@ -194,6 +225,7 @@ instance ToField LT.Text where
     toField = toField . LT.toStrict
     {-# INLINE toField #-}
 
+
 instance ToField UTCTime where
     toField = Plain . inQuotes . utcTimeToBuilder
     {-# INLINE toField #-}
@@ -230,6 +262,10 @@ instance ToField Date where
     toField = Plain . inQuotes . dateToBuilder
     {-# INLINE toField #-}
 
+instance ToField NominalDiffTime where
+    toField = Plain . inQuotes . nominalDiffTimeToBuilder
+    {-# INLINE toField #-}
+
 instance (ToField a) => ToField (PGArray a) where
     toField xs = Many $
         Plain (fromByteString "ARRAY[") :
@@ -262,3 +298,66 @@ toJSONField = toField . JSON.toJSON
 inQuotes :: Builder -> Builder
 inQuotes b = quote `mappend` b `mappend` quote
   where quote = Utf8.fromChar '\''
+
+interleaveFoldr :: (a -> [b] -> [b]) -> b -> [b] -> [a] -> [b]
+interleaveFoldr f b bs as = foldr (\a bs -> b : f a bs) bs as
+{-# INLINE interleaveFoldr #-}
+
+instance ToRow a => ToField (Values a) where
+    toField (Values types rows) =
+        case rows of
+          []    -> case types of
+                     []    -> error norows
+                     (_:_) -> values $ typedRow (repeat (lit "null"))
+                                                types
+                                                [lit " LIMIT 0)"]
+          (_:_) -> case types of
+                     []    -> values $ untypedRows rows [litC ')']
+                     (_:_) -> values $ typedRows rows types [litC ')']
+      where
+        funcname = "Database.PostgreSQL.Simple.toField :: Values a -> Action"
+        norows   = funcname ++ "  either values or types must be non-empty"
+        emptyrow = funcname ++ "  each row must contain at least one column"
+        lit  = Plain . fromByteString
+        litC = Plain . fromChar
+        values x = Many (lit "(VALUES ": x)
+
+        typedField :: (Action, QualifiedIdentifier) -> [Action] -> [Action]
+        typedField (val,typ) rest = val : lit "::" : toField typ : rest
+
+        typedRow :: [Action] -> [QualifiedIdentifier] -> [Action] -> [Action]
+        typedRow (val:vals) (typ:typs) rest =
+            litC '(' :
+              typedField (val,typ) ( interleaveFoldr
+                                        typedField
+                                        (litC ',')
+                                        (litC ')' : rest)
+                                        (zip vals typs)   )
+        typedRow _ _ _ = error emptyrow
+
+        untypedRow :: [Action] -> [Action] -> [Action]
+        untypedRow (val:vals) rest =
+            litC '(' : val :
+            interleaveFoldr
+                 (:)
+                 (litC ',')
+                 (litC ')' : rest)
+                 vals
+        untypedRow _ _ = error emptyrow
+
+        typedRows :: ToRow a => [a] -> [QualifiedIdentifier] -> [Action] -> [Action]
+        typedRows [] _ _ = error funcname
+        typedRows (val:vals) types rest =
+            typedRow (toRow val) types (multiRows vals rest)
+
+        untypedRows :: ToRow a => [a] -> [Action] -> [Action]
+        untypedRows [] _ = error funcname
+        untypedRows (val:vals) rest =
+            untypedRow (toRow val) (multiRows vals rest)
+
+        multiRows :: ToRow a => [a] -> [Action] -> [Action]
+        multiRows vals rest = interleaveFoldr
+                                (untypedRow . toRow)
+                                (litC ',')
+                                rest
+                                vals
