@@ -3,7 +3,9 @@
 {-# LANGUAGE ScopedTypeVariables #-}
 import Common
 import Database.PostgreSQL.Simple.FromField (FromField)
+import Database.PostgreSQL.Simple.Types(Query(..),Values(..))
 import Database.PostgreSQL.Simple.HStore
+import Database.PostgreSQL.Simple.Copy
 import qualified Database.PostgreSQL.Simple.Transaction as ST
 import Control.Applicative
 import Control.Exception as E
@@ -13,8 +15,10 @@ import Data.IORef
 import Data.Typeable
 import qualified Data.ByteString as B
 import Data.Map (Map)
+import Data.List (sort)
 import qualified Data.Map as Map
 import Data.Text(Text)
+import qualified Data.Text.Encoding as T
 import System.Exit (exitFailure)
 import System.IO
 import qualified Data.Vector as V
@@ -36,6 +40,10 @@ tests =
     , TestLabel "HStore"        . testHStore
     , TestLabel "JSON"          . testJSON
     , TestLabel "Savepoint"     . testSavepoint
+    , TestLabel "Unicode"       . testUnicode
+    , TestLabel "Values"        . testValues
+    , TestLabel "Copy"          . testCopy
+    , TestLabel "Double"        . testDouble
     ]
 
 testBytea :: TestEnv -> Test
@@ -159,6 +167,7 @@ testArray TestEnv{..} = TestCase $ do
 
 testHStore :: TestEnv -> Test
 testHStore TestEnv{..} = TestCase $ do
+    execute_ conn "CREATE EXTENSION IF NOT EXISTS hstore"
     roundTrip []
     roundTrip [("foo","bar"),("bar","baz"),("baz","hello")]
     roundTrip [("fo\"o","bar"),("b\\ar","baz"),("baz","\"value\\with\"escapes")]
@@ -243,6 +252,72 @@ testSavepoint TestEnv{..} = TestCase $ do
 
     return ()
 
+testUnicode :: TestEnv -> Test
+testUnicode TestEnv{..} = TestCase $ do
+    let q = Query . T.encodeUtf8  -- Handle encoding ourselves to ensure
+                                  -- the table gets created correctly.
+    let messages = map Only ["привет","мир"] :: [Only Text]
+    execute_ conn (q "CREATE TEMPORARY TABLE ру́сский (сообщение TEXT)")
+    executeMany conn "INSERT INTO ру́сский (сообщение) VALUES (?)" messages
+    messages' <- query_ conn "SELECT сообщение FROM ру́сский"
+    sort messages @?= sort messages'
+
+testValues :: TestEnv -> Test
+testValues TestEnv{..} = TestCase $ do
+    execute_ conn "CREATE TEMPORARY TABLE values_test (x int, y text)"
+    test (Values ["int4","text"] [])
+    test (Values ["int4","text"] [(1,"hello")])
+    test (Values ["int4","text"] [(1,"hello"),(2,"world")])
+    test (Values ["int4","text"] [(1,"hello"),(2,"world"),(3,"goodbye")])
+    test (Values [] [(1,"hello")])
+    test (Values [] [(1,"hello"),(2,"world")])
+    test (Values [] [(1,"hello"),(2,"world"),(3,"goodbye")])
+  where
+    test :: Values (Int, Text) -> Assertion
+    test table@(Values _ vals) = do
+      execute conn "INSERT INTO values_test ?" (Only table)
+      vals' <- query_  conn "DELETE FROM values_test RETURNING *"
+      sort vals @?= sort vals'
+
+
+testCopy :: TestEnv -> Test
+testCopy TestEnv{..} = TestCase $ do
+    execute_ conn "CREATE TEMPORARY TABLE copy_test (x int, y text)"
+    copy_ conn "COPY copy_test FROM STDIN (FORMAT CSV)"
+    mapM_ (putCopyData conn) copyRows
+    putCopyEnd conn
+    copy_ conn "COPY copy_test FROM STDIN (FORMAT CSV)"
+    mapM_ (putCopyData conn) abortRows
+    putCopyError conn "aborted"
+    -- Hmm, does postgres always produce \n as an end-of-line here, or
+    -- are there cases where it will use a \r\n as well?
+    copy_ conn "COPY copy_test TO STDOUT (FORMAT CSV)"
+    rows <- loop []
+    sort rows @?= sort copyRows
+    -- Now, let's just verify that the connection state is back to ready,
+    -- so that we can issue more queries:
+    [Only (x::Int)] <- query_ conn "SELECT 2 + 2"
+    x @?= 4
+  where
+    copyRows  = ["1,foo\n"
+                ,"2,bar\n"]
+    abortRows = ["3,baz\n"]
+    loop rows = do
+      mrow <- getCopyData conn
+      case mrow of
+        CopyOutDone _   -> return rows
+        CopyOutRow  row -> loop (row:rows)
+
+testDouble :: TestEnv -> Test
+testDouble TestEnv{..} = TestCase $ do
+    [Only (x :: Double)] <- query_ conn "SELECT 'NaN'::float8"
+    assertBool "expected NaN" (isNaN x)
+    [Only (x :: Double)] <- query_ conn "SELECT 'Infinity'::float8"
+    x @?= (1 / 0)
+    [Only (x :: Double)] <- query_ conn "SELECT '-Infinity'::float8"
+    x @?= (-1 / 0)
+
+
 data TestException
   = TestException
   deriving (Eq, Show, Typeable)
